import os
import sys
import re
import os.path
import numpy as np
import pickle

if (sys.platform == 'win32'):
    g_basedir = r'\\datastore\valhalla\instruments\instrument_studies\aim-north\2019-03-01-sensitivity_study'
else:
    g_basedir = '/datastore/valhalla/instruments/instrument_studies/aim-north/2019-03-01-sensitivity_study'
#   g_basedir = '.'


#------------------------------------------------------------------------------
#           Level1_AimNorth
#------------------------------------------------------------------------------

class AimNorth_NO2Level1:
    """
    A class used to read in Aim North NO2 Level 1 products generated by the simulation code.

    Attributes
    ----------
    radiance : np.ndarray [wavelen, lats, hours]
        Three dimensional array of radiance at the front aperture.  Returns 'mean' convolved signal across wavelength bandpass and mean signal in solid angle

    toa_albedo : np.ndarray [wavelen, lats, hours]
        Three dimesnional array of top-of-atmosphere albedo at front aperture

    irradiance : np.ndarray [wavelen, lats, hours]
        Three dimesnional array of irradiance at front aperture. Photons /cm2. Signal includes width of the bandpass filter and the size of the solid angle

    photon_flux : np.ndarray [wavelen, lats, hours]
        Three dimensional array of photon_flux passing through the front aperture. This radiance multiplied by the area of the front aperture.

    look_azimuth : np.ndarray [lats, hours]
        The geographic azimuth of the look vector at each ground point

    look_zenith : np.ndarray [lats, hours]
        The zenith angle of the look vector at each ground point

    ground_saa : np.ndarray [lats, hours]
        The solar azimuth angle at each ground point

    ground_ssa : np.ndarray [lats, hours]
        The solar scattering angle at each ground point.

    ground_sza : np.ndarray [lats, hours]
        The solar zenith angle at each ground point

    sun_vector : np.ndarray [ (x,y,z), hours]
        The geocentric unit vector from the center of the earth to the sun.

    satellite_geocentric_los : np.ndarray [ (x,y,z), lats, hours]
        The look vector from the satellite towards each ground point in geocentric coords

    satellite_geocentric_pos : np.ndarray [(x,y,z), lats, hours]
        The geocentric position of the satellite when looking at each ground point.

    satellite_geodetic_pos : np.ndarray [(lat,lng,height), lats, hours]
        The geodetic coordinates of the satellite when looking at each ground point.

    ground_points_geodetic_pos : np.ndarray  [(lat,lng,h), lats, hours]
        The geodetic location of each ground point.

    no2_profile : np.ndarray [heights, hours]
        The "true" NO2 number density profile used in the simulation. It is specified in molecules/cm3 at each height in `profileheights`. The same profile
        is used for all ground points at the same instant in time.

    airnumberdensity : np.ndarray [heights, hours]
        The air number density profile used in the simulation. It is specified in molecules/cm3 at each height in `profileheights`. The same profile
        is used for all ground points at the same instant in time.

    temperature : np.ndarray  [heights, hours]
        The temperature (Kelvin) profile used in the simulation. It is specified in Kelvin at each height in `profileheights`. The same profile
        is used for all ground points at the same instant in time.

    profileheights : np.ndarray [heights, hours]
        The heights used to specify profiles used in the simulation.

    utc : np.ndarray [hours] of  np.datetime64
        The array of times used in the simulation. Typically 12 times are scalculated for each day. Each time is specified as a numpy datetime64 instance.

    version : np.ndarray [hours]
        An array of strings specifying the Level 1 simulator version used to generate the signals.

    wavelengths : np.ndarray [wavelen]
        The array of wavelength in nanometers used in this simulation

    latitudes : np.ndarray [lats]
        The array of latitudes in degrees used in this simulation

    """

    #------------------------------------------------------------------------------
    #           __init__
    #------------------------------------------------------------------------------

    def __init__(self):

        self.radiance                   : np.ndarray = None     # [wavelen, lats, hours]
        self.toa_albedo                 : np.ndarray = None     # [wavelen, lats, hours]
        self.irradiance                 : np.ndarray = None     # [wavelen, lats, hours]
        self.photon_flux                : np.ndarray = None     # [wavelen, lats, hours]

        self.look_azimuth               : np.ndarray = None     # [lats, hours]
        self.look_zenith                : np.ndarray = None     # [lats, hours]
        self.ground_saa                 : np.ndarray = None     # [lats, hours]
        self.ground_ssa                 : np.ndarray = None     # [lats, hours]
        self.ground_sza                 : np.ndarray = None     # [lats, hours]
        self.sun_vector                 : np.ndarray = None     # [ (x,y,z), hours]

        self.satellite_geocentric_los   : np.ndarray = None     # [ (x,y,z), lats, hours]
        self.satellite_geocentric_pos   : np.ndarray = None     # [(x,y,z), lats, hours]
        self.satellite_geodetic_pos     : np.ndarray = None     # [(lat,lng,height), lats, hours]
        self.ground_points_geodetic_pos : np.ndarray = None     # [(lat,lng,h), lats, hours]

        self.no2_profile                : np.ndarray = None     # [heights, hours]
        self.airnumberdensity           : np.ndarray = None     # [heights, hours]
        self.temperature                : np.ndarray = None     # [heights, hours]
        self.profileheights             : np.ndarray = None     # [heights, hours]

        self.utc                        : np.ndarray = None     # [hours] array of np.datetime64
        self.version                    : np.ndarray = None     # [hours]
        self.wavelengths                : np.ndarray = None     # [wavelen]
        self.latitudes                  : np.ndarray = None     # [lats]


    #------------------------------------------------------------------------------
    #           _load_radiance_pickle_file
    #------------------------------------------------------------------------------

    def _load_radiance_pickle_file(self, filename: str):
        with open(filename, 'rb') as f:
            data = pickle.load(f)
        return data

    #------------------------------------------------------------------------------
    #           load_files
    #------------------------------------------------------------------------------

    def load_data( self,
                    no2profile_type : str,
                    psf             : float,
                    albedo          : float,
                    year_month      : str):

        dirname = g_basedir +os.sep + 'radiance_calculations' + os.sep + 'psf_{:1.1f}'.format(psf) + os.sep+ '{}'.format(no2profile_type)
        filespec = r'aimnorth_{}_{:1.1f}_{:1.1f}_{}-21-\d\d-00-00.pickle'.format(no2profile_type, psf, albedo,year_month)
        files = os.listdir(dirname)
        goodfiles = []
        for name in files:
            m = re.match(filespec, name)
            if (m is not None):
                goodfiles.append(name)
        goodfiles.sort()
        alldata = None

        nt = len(goodfiles)
        for i in range(nt):
            filename = dirname + os.sep + goodfiles[i]
            data = self._load_radiance_pickle_file(filename)
            if (alldata is None):                                                   # If this is the first time through
                alldata = {}                                                        # then create the dictionar
                for key,value in data.items():
                    shape = None
                    if type(value) is np.ndarray:
                        shape = value.shape
                        dtype = value.dtype
                    else:
                        shape = [1]
                        dtype = None
                        if type(value) is str:           dtype = '<U10'
                        if type(value) is np.datetime64: dtype = 'datetime64[us]'

                    newshape = list(shape)
                    newshape.append(nt)
                    alldata[key] = np.empty( newshape, dtype=dtype)
            for key, value in data.items():
                 alldata[key][...,i] = value

        self.radiance                   = alldata['radiance']
        self.toa_albedo                 = alldata['toa_albedo']
        self.irradiance                 = alldata['irradiance']
        self.photon_flux                = alldata['photon_flux']
        self.look_azimuth               = alldata['look_azimuth']
        self.look_zenith                = alldata['look_zenith']
        self.ground_saa                 = alldata['ground_saa']
        self.ground_ssa                 = alldata['ground_ssa']
        self.ground_sza                 = alldata['ground_sza']
        self.sun_vector                 = alldata['sun_vector']
        self.satellite_geocentric_los   = alldata['satellite_geocentric_los']
        self.satellite_geocentric_pos   = alldata['satellite_geocentric_pos']
        self.satellite_geodetic_pos     = alldata['satellite_geodetic_pos']
        self.ground_points_geodetic_pos = alldata['ground_points_geodetic_pos']
        self.no2_profile                = alldata['no2_profile']
        self.airnumberdensity           = alldata['airnumberdensity']
        self.temperature                = alldata['temperature']
        self.profileheights             = alldata['profileheights']
        self.utc                        = np.squeeze( alldata['utc'] )
        self.version                    = alldata['version']
        self.wavelengths                = alldata['wavelengths']
        self.latitudes                  = np.arange(40, 80)




#------------------------------------------------------------------------------
#           AimNorth_NO2Level2
#------------------------------------------------------------------------------

class AimNorth_NO2Level2:
    """
    A class used to read in Aim North NO2 Level 2 products generated by the simulation code.

    Attributes
    ----------
    snr: float
        The SNR ratio used when generating this file

    utc: np.ndarray [hours]
        An array of numpy.datetime64 that have the utc times of the level 2 retrievals.  It is usually on the 12 even UTC hours of each day

    psf: float
        The point spread function (FWHM) of the AIM north spectrometer used to generate this data set.

    albedo: float
        The albedo used to generate this data set.

    profile_type: str
        The type of NO2 profile. Typically one of 'clean', 'urban' or 'oilsand'. The same NO2 profile is used for all latitudes and times in one data set

    vcd: np.ndarray [lats, random samples, hours]
        Three dimensional array of vertical column density. The random samples are taken by applying the SNR to the level 1 toa_albedo signal and randomly sampling from the distribution before doing the retrieval.

    vcd_error: np.ndarray [lats, random samples, hours]
        Three dimension array of errors on the vertical column density as calculated by the retrieval code.

    vcd_avg: np.ndarray [lats, hours]
        The average vertical column density for each latitude and time point taken from the mean of the random sampled `vcd` array.

    vcd_std: np.ndarray [lats, hours]
        The standard deviation of `vcd_avg` for each latitude and time point. Should be comparable to `vcd_error_avg` if all things are equal.

    vcd_error_avg: np.ndarray [lats, hours]
        The average value of `vcd_error` for each latitude and time point

    vcd_error_std: np.ndarray = None
        The standard deviation of `vcd_error` for each latitude and time point.

    scd: np.ndarray [lats, random samples, hours]
        Three dimensional array of slant column density. The random samples are taken by applying the SNR to the level 1 toa_albedo signal and randomly sampling from the distribution before doing the retrieval.

    scd_error: np.ndarray [last, randon samples, hours]
        Three dimension array of errors on the slant  column density as calculated by the retrieval code.

    amf: np.ndarray [ lats, random samples, hours]
        Three dimensional array of air mass factors. The random samples are provided but are all identical to each other for a given latitude and hour.

    amf_error: np.ndarray [ lats, random samples, hours]
        Three dimensional array of air mass factors errors. These are the values calculated by the retrieval code.

    true_shells: np.ndarray [heights]
        A 1d array specifying the height in meters where the true NO2 profile is provided. Typicall provided at 500 m intervals.

    true_no2profile: np.ndarray [heights]
        A 1d array specifying the height profile of the NO2 number density (molecules/cm3) on the height grid `true_shells`

    true_vcd: float
        The true NO2 vertical column density used in this simulation.

    latitudes: np.ndarray [lats]
        The array of latitudes used in this simulation
    """

    #------------------------------------------------------------------------------
    #           __init__
    #------------------------------------------------------------------------------

    def __init__(self):
        self.snr             : float      = None
        self.utc             : np.ndarray = None
        self.psf             : float      = None
        self.albedo          : float      = None
        self.profile_type    : str        = None
        self.vcd             : np.ndarray = None
        self.vcd_error       : np.ndarray = None
        self.vcd_avg         : np.ndarray = None
        self.vcd_std         : np.ndarray = None
        self.vcd_error_avg   : np.ndarray = None
        self.vcd_error_std   : np.ndarray = None
        self.scd             : np.ndarray = None
        self.scd_error       : np.ndarray = None
        self.amf             : np.ndarray = None
        self.amf_error       : np.ndarray = None
        self.true_shells     : np.ndarray = None
        self.true_no2profile : np.ndarray = None
        self.true_vcd        : float      = None
        self.latitudes       : np.ndarray = None


    #------------------------------------------------------------------------------
    #           make_filename
    #------------------------------------------------------------------------------

    def _make_filename(self, profile_type, psf, albedo, date_month):

        filename = g_basedir + os.sep + 'retrieval_stats' + os.sep + 'psf_{:1.1f}'.format(psf) + os.sep + profile_type + os.sep +'aimnorth_retrieval_{}_{:1.1f}_{:1.1f}_{}-21.pickle'.format(profile_type, psf,albedo,date_month)
        return filename

    #------------------------------------------------------------------------------
    #           load_retrieval_file
    #------------------------------------------------------------------------------

    def load_data( self, profile_type, psf, albedo, date_month, snr : int ):

        filename =  self._make_filename( profile_type, psf, albedo, date_month )
        with open(filename, 'rb') as f:  # and save the results to a pickle file
            answers = pickle.load( f )

            a = answers[snr]
            utc = []
            for h in a['hours']:
                utc.append( '{}-21T{:02d}:00:00.000000'.format(a['day'], h))

            self.utc             = np.array( utc, dtype = 'datetime64')
            self.snr             = float(a['snr'])
            self.psf             = a['psf']
            self.albedo          = a['albedo']
            self.profile_type    = a['profile_type']                     # urban, clean, oilsand
            self.vcd             = a['vcd']                              # [latitudes, random_samples, hours]
            self.vcd_error       = a['vcd_error']                        # [latitudes, random_samples, hours]
            self.vcd_avg         = np.mean(a['vcd'], axis = 1 )          # [latitudes, hours]
            self.vcd_std         = np.std (a['vcd'], axis = 1 )          # [latitudes, hours]
            self.vcd_error_avg   = np.mean(a['vcd_error'], axis=1)       # [latitudes, hours]
            self.vcd_error_std   = np.std(a['vcd_error'], axis=1)        # [latitudes, hours]
            self.scd             = a['scd']                              # [latitudes, random_samples, hours]
            self.scd_error       = a['scd_error']                        # [latitudes, random_samples, hours]
            self.amf             = a['amf']                              # [latitudes, random_samples, hours]
            self.amf_error       = a['amf_error']                        # [latitudes, random_samples, hours]
            self.true_shells     = a['true_shells']                      # [numheights] Shell heights of true_profile
            self.true_no2profile = a['true_no2profile']                  # [numheights] The true NO2 profile
            self.true_vcd        = a['true_vcd']                         # True NO2 vertical column
            self.latitudes       = np.arange(40,80)                      # [numlatitudes] The latitudes



if __name__ == "__main__":

    L1 = AimNorth_NO2Level1()
    L2 = AimNorth_NO2Level2()
    L1.load_data( 'urban', 0.4, 0.3, '2019-06')
    L2.load_data( 'urban', 0.4, 0.3, '2019-06', snr = 700)
    print('Finished experiment')
