import os
import os.path
import sys
import argparse
import numpy as np
import pickle
from aimnorthsimulator.aimnorth_l1 import AimNorth_Simulator
import sasktran
import scipy.integrate
#------------------------------------------------------------------------------
#           load_coarse_no2_profile
#------------------------------------------------------------------------------

def load_coarse_no2_profile(profile, boundary_layer_width, layer_width):
    """
    Loads in a coarse version of a fine structured NO2 profile. The coarse version is adjusted so it has the same total
    column NO2 as the fine structure. This redistribuition is done so the linear interpolatioon of sasktran between
    shells is consistent with the retrievals etc.

     """

    map = {'urban': 0, 'oilsand': 1, 'clean': 2}
    profile_idx = map[profile]

    dirname,filename = os.path.split(__file__)
    if (dirname is not None) and (len(dirname) > 0):
        dirname += os.sep
    else:
        dirname = ""

    with  open( dirname + 'toronto_oilsands_north_no2.pkl', 'rb') as f:
        data = pickle.load(f)

    no2 = data['no2_cm3'][profile_idx, :]
    alt = data['altitude'][profile_idx, :]


    vcd = scipy.integrate.trapz(no2, 1e2 * alt)

    h = np.linspace(0, 8e4, 1601)
    v = np.interp(h, alt, no2)
    hgrid = np.arange( 0.0, 8e4 + 1, layer_width)
    hgrid = np.append(np.arange(0, 2000., boundary_layer_width), np.arange(2000., 8e4 + 1, layer_width))
    vgrid = np.zeros_like(hgrid)
    vgrid[0] = v[0]
    vgrid[-1] = v[-1]
    for j in range(1, hgrid.size-1):
        h0 = hgrid[j-1]
        h1 = hgrid[j]
        h2 = hgrid[j+1]
        idx = np.where( (h >= 0.5*(h0+h1)) & (h <= 0.5*(h1+h2) ) )
        vgrid[j] = np.average( v[idx] )

    vcd_coarse = scipy.integrate.trapz(vgrid, 1e2*hgrid)

    dv = vcd_coarse - vcd
    vgrid[0] -= 2e-2 * dv / (hgrid[1] - hgrid[0])

    vcd_coarse = scipy.integrate.trapz(vgrid, 1e2*hgrid)
    if np.abs(vcd - vcd_coarse) > 10.:
        raise ValueError(f'VCDs do not match: original {vcd} coarse {vcd_coarse}')
    no2profile = sasktran.ClimatologyUserDefined(hgrid, {'SKCLIMATOLOGY_NO2_CM3': vgrid})
    return no2profile

#------------------------------------------------------------------------------
#           def load_original_no2_profile(profile):
#------------------------------------------------------------------------------

def load_original_no2_profile(profile):
    with open('Z:\\luf542\\G5NR_HERACLES\\toronto_oilsands_north_no2.pkl', 'rb') as f:
        data = pickle.load(f)
    map = {'urban': 0, 'oilsand': 1, 'clean': 2}
    alt = data['altitude'][map[profile], :]
    no2 = data['no2_cm3'][map[profile], :]
    return no2, alt


#------------------------------------------------------------------------------
#           load_no2_profile
#------------------------------------------------------------------------------

def load_test_no2_profile(index)->sasktran.Climatology:
    """
    Fetches one of three typical NO2 profiles.
     0 = Urban
     1 = Oilsand
     2 = Clean prairie
     The python pickle file was generated by Lukas Fehr from the GeosChem 5 model.

     This is now deprecated as it produces NO2 on too fine a grid and we get inconsistencies in the retrieval
     when Sasktran interpolates on a 1 km grid to the 500 m grod of teh NO2 profile. Use `load_coarse_no2_profile`

    """
    dirname,filename = os.path.split(__file__)
    if (dirname is not None) and (len(dirname) > 0):
        dirname += os.sep
    else:
        dirname = ""
    f = open( dirname + 'toronto_oilsands_north_no2.pkl', 'rb')
    q = pickle.load(f)
    f.close()
    heights = q['altitude'][index,:]
    values  = q['no2_cm3'][index,:]

    # ---- INterpolate onto a super fine grid
    h = np.arange( 0, heights[-1], 50.0)
    v = np.interp( h, heights,values, right = 1.0)

    # and then get the average profile on a 500 meter grid

    hgrid = np.arange( 0.0, heights[-1]+499.0, 500.0)
    vgrid = np.zeros_like(hgrid)
    vgrid[0] = values[0]
    vgrid[-1] = values[-1]
    for i in range(1, hgrid.size-1):
        h0 = hgrid[i-1]
        h1 = hgrid[i]
        h2 = hgrid[i+1]
        idx = np.where( (h >= 0.5*(h0+h1)) & (h <= 0.5*(h1+h2) ) )
        vgrid[i] = np.average( v[idx] )

    no2profile = sasktran.ClimatologyUserDefined(hgrid, {'SKCLIMATOLOGY_NO2_CM3': vgrid})
    return no2profile

#------------------------------------------------------------------------------
#           make_atmosphere
#------------------------------------------------------------------------------

def make_atmosphere( no2profile_type: str, albedo, wavelengths:np.ndarray, psf: float)->sasktran.Atmosphere:

    atmosphere = sasktran.Atmosphere()

    if (no2profile_type == 'test1'):
        atmosphere.atmospheric_state = sasktran.MSIS90()
        atmosphere.brdf              = albedo
        atmosphere['air']            = sasktran.Species(sasktran.Rayleigh(), sasktran.MSIS90())
        atmosphere['ozone']          = sasktran.Species(sasktran.O3DBM(), sasktran.Labow())
        atmosphere['no2']            = sasktran.Species(sasktran.NO2Vandaele1998(), sasktran.Pratmo())

    if (no2profile_type == 'test2'):
        atmosphere.atmospheric_state = sasktran.MSIS90()
        atmosphere.brdf              = albedo
        atmosphere['air']            = sasktran.Species(sasktran.Rayleigh(), sasktran.MSIS90())
        atmosphere['ozone']          = sasktran.Species(sasktran.O3DBM(), sasktran.Labow())
        atmosphere['no2']            = sasktran.Species(sasktran.NO2Vandaele1998(), sasktran.Pratmo())

    else:
        msis =  sasktran.MSIS90()
        no2profile                   = load_coarse_no2_profile( no2profile_type, 1000.0, 1000.0 )
        atmosphere.atmospheric_state = msis
        atmosphere.brdf              = albedo
        atmosphere['air']            = sasktran.Species(sasktran.Rayleigh(), msis)
        atmosphere['ozone']          = sasktran.Species(sasktran.OpticalPropertyConvolved( sasktran.O3DBM(),           psf_wavelength=wavelengths,  psf=psf/2.3548200450309493820231386529194), sasktran.Labow())
        atmosphere['no2'  ]          = sasktran.Species(sasktran.OpticalPropertyConvolved( sasktran.NO2Vandaele1998(), psf_wavelength=wavelengths,  psf=psf/2.3548200450309493820231386529194), no2profile)
    return atmosphere

#------------------------------------------------------------------------------
#           test_aimnorth
#------------------------------------------------------------------------------

def experiment_psf_variation( do_lukas_test = 0):

# syncrude at 57.04,     -111.60
# Toronto at  43.78175N,  -79.4696163N
# Saskatoon at 52.1315,  -106.6335
# Loop over point spread function
# acquire dirty NO2 over Toronto and oil sand.
# acquire clean NO2 over Saskatoon.
# add reference point to output file.
# Add NO2 profile 500m spacing
# Add air profile on same grid
# Add Temperature profile on same grid.

    if (sys.platform =='win32'):
        basedir    = r'\\datastore\valhalla\instruments\instrument_studies\aim-north\2019-03-01-sensitivity_study\radiance_calculations'
    else:
        #basedir = '/datastore/valhalla/instruments/instrument_studies/aim-north/2019-03-01-sensitivity_study/radiance_calculations'
        basedir = 'radiance_calculations'

    if (do_lukas_test == "test1"):
        no2profile_array = ['test1']
        dates       = ['2006-09-22']
        hours       = [14]
        minutes     = 24
        psfarray    = [0.4]
        albedoarray = [0.3]

        slat, slon = 52.1, -106.7  # Saskatoon
        tlat, tlon = 43.7, -79.4  # Toronto
        ilat, ilon = 63.8, -68.5  # Iqaluit
        ylat, ylon = 62.5, -114.4  # Yellowknife
        latitudes  = np.array( [slat, tlat, ilat, ylat])
        longitudes = np.array( [slon, tlon, ilon, ylon])
        center_lat  = 60
        center_long = -90
        center_height  = 35786e3
        wavelengths = np.arange(405., 465., 0.3)

    if (do_lukas_test == "test2"):
        no2profile_array = ['test2']
        dates       = ['2006-09-22']
        hours       = [14]
        minutes     = 24
        psfarray    = [0.2, 0.4,0.6,0.8,1.0]
        albedoarray = [0.3]

        slat, slon = 52.1, -106.7  # Saskatoon
        tlat, tlon = 43.7, -79.4  # Toronto
        ilat, ilon = 63.8, -68.5  # Iqaluit
        ylat, ylon = 62.5, -114.4  # Yellowknife
        latitudes  = np.array( [slat, tlat, ilat, ylat])
        longitudes = np.array( [slon, tlon, ilon, ylon])
        center_lat  = 60
        center_long = -90
        center_height  = 35786e3
        wavelengths = np.arange(405., 465., 0.3)

    else:
        lats       = np.arange(40,80,1)                     # Create the latitudes  of the points we want to look at
        #lons       = np.array ([-80.0, -95.0, -110.0])      # Create the longitudes of the points we want to look at
        lons       = np.array( [-80] )
        latitudes  = np.repeat( lats, lons.size)
        longitudes = np.repeat( lons, lats.size)
        center_lat  = 60
        center_long = -95
        center_height = 37770000.0
        no2profile_array    = ['clean','urban','oilsand']
        dates               = ['2018-12-21', '2019-03-21', '2019-06-21', '2019-09-21']
        hours               = [0,2,4,6,8,10,12,14,16,18,20,22]
        minutes             = 0
        psfarray            = [0.2,0.4,0.6,0.8,1.0]
        albedoarray         = [0.0, 1.0, 0.3]
        #wavelengths        = np.arange( 320,800, 0.2)
        wavelengths         = np.arange( 400,470, 0.2)

    sasktran_options   = None if (sys.platform == 'win32') else {'num_threads': 32}
    sasktranif_options = None

    for no2profile_type in no2profile_array:
        for psf in psfarray:
            for albedo in albedoarray:
                for d in dates:
                    for h in hours:
                        utcstr  = '{}T{:02d}:{:02d}:00.000000'.format(d,h,minutes)
                        aimnorth = AimNorth_Simulator()                                                                         # Create the simulator object. This will load default settings from the internal YAML file
                        aimnorth.instrument_settings['wavelengths'] = wavelengths                                               # set the wavelengths
                        aimnorth.instrument_settings['psf']         = psf                                                       # set the point spread function
                        outputdir = basedir + os.sep + 'psf_{:1.1f}'.format(psf) + os.sep+ no2profile_type
                        try:
                            os.makedirs(outputdir, exist_ok=True)
                        except:
                            print("Unable to create directory {}:".format(outputdir), sys.exc_info()[0])

                        filename  = outputdir + os.sep + 'aimnorth_{}_{:1.1f}_{:1.1f}_{}-{:02d}-00-00.pickle'.format(no2profile_type,psf,albedo, d,h)
                        if not os.path.exists(filename):
                            print("Processing {}, psf ={:1.1f}, albedo = {:1.1f} for {}".format(no2profile_type, psf, albedo, utcstr))
                            atmosphere                   = make_atmosphere( no2profile_type, albedo,
                                                                            aimnorth.instrument_settings['wavelengths'],
                                                                            aimnorth.instrument_settings['psf'] )
                            aimnorth.set_platform_location         ( center_lat, center_long, np.datetime64(utcstr), height_meters=center_height)         # set the time and location of the satellite
                            aimnorth.set_ground_target_locations   ( latitudes, longitudes)                                             # set the location of the ground target points
                            level1   = aimnorth.calculate_radiance ( atmosphere, sasktran_options = sasktran_options, sasktranif_options = sasktranif_options)                                                               # calculate the radiance using the current settings
                            try:
                                with open(filename, 'wb') as f:                                                                             # and save the results to a pickle file
                                    pickle.dump(level1, f, pickle.HIGHEST_PROTOCOL)
                            except:
                                print("Unexpected file I/O error:", sys.exc_info()[0])

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--test", help="run Lukas Fehr test case", action="store")
    args = parser.parse_args()
    if ( args.test is None): do_lukas_test = None
    else:                    do_lukas_test = "test"+args.test
    experiment_psf_variation(do_lukas_test=do_lukas_test)
    print('Finished')
