from typing import List
import os
import os.path
import sys
import pickle
import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate

from skdoas.retrieval.lineshape          import Gaussian
from skdoas.retrieval.omno2              import CrossSectionOMNO2,SolarSpectrumOMNO2,RingSpectrumOMNO2,SlantColumnOMNO2,RetrievalOMNO2
from aimnorthretrieval.retrieval_objects import Radiance_AimNorthStudyMar2019, Atmosphere_AimNorthStudyMar2019, AMF_AimNorthStudyMar2019, AMFLookup_AimNorthStudyMar2019

#------------------------------------------------------------------------------
#           make_path
#------------------------------------------------------------------------------

def  make_path(psf=0.4, profile='clean', albedo=0.3, year_month='2019_06', hour=18, old=False):

    psfstr    = '{:1.1f}'.format(psf)
    albedostr = '{:1.1f}'.format(albedo)
    hourstr   = '{:02d}'.format(int(hour))
    if old:
        if (sys.platform == 'win32'):
            basedir = r'\\datastore\valhalla\instruments\instrument_studies\aim-north\2019-03-01-sensitivity_study\radiance_calculations_old'
        else:
            basedir =   '/datastore/valhalla/instruments/instrument_studies/aim-north/2019-03-01-sensitivity_study/radiance_calculations_old'
    else:
        if (sys.platform == 'win32'):
            basedir = r'\\datastore\valhalla\instruments\instrument_studies\aim-north\2019-03-01-sensitivity_study\radiance_calculations'
        else:
            basedir =   '/datastore/valhalla/instruments/instrument_studies/aim-north/2019-03-01-sensitivity_study/radiance_calculations'

    basedir += os.sep
    filepath = os.path.join(basedir, f'psf_{psfstr}{os.sep}{profile}{os.sep}aimnorth_{profile}_{psfstr}_{albedostr}_{year_month}-21-{hourstr}-00-00.pickle')
    return filepath

#------------------------------------------------------------------------------
#           make_path
#------------------------------------------------------------------------------

def  make_stats_path(psf=0.4, profile='clean', albedo=0.3, year_month='2019_06', old=False):

    psfstr    = '{:1.1f}'.format(psf)
    albedostr = '{:1.1f}'.format(albedo)

    if (sys.platform =='win32'):
        basedir    = r'\\datastore\valhalla\instruments\instrument_studies\aim-north\2019-03-01-sensitivity_study\retrieval_stats'
    else:
        basedir = '/datastore/valhalla/instruments/instrument_studies/aim-north/2019-03-01-sensitivity_study/retrieval_stats'

    basedir += os.sep
    filepath = os.path.join(basedir, f'psf_{psfstr}{os.sep}{profile}{os.sep}aimnorth_retrieval_{profile}_{psfstr}_{albedostr}_{year_month}-21.pickle')
    return filepath

#------------------------------------------------------------------------------
#           do_retrievals
#------------------------------------------------------------------------------

def do_retrievals(  profile    : str,
                    psf        : float,
                    albedo     : float,
                    year_month : str,
                    hours      : List[int],
                    snr_range  : List[float],
                    numsamples : int):
    """
    Executes the NO2 retrieval for a given test case and generates statistics on the accuracy/variance of the retrieval.
    The retrieval is performed `numsamples` times for each signal-to-noise ratio in `snr_range` by adding random normal
    noise to the calculated radiances at the specified snr. The set of retrievals
    is performed for each the hours in the given test case scenario.

    The choice of test case is not arbitrary and must be selected from one of the actual aim-north test cases.

    Writes the retreived

    Parameters
    ----------
    profile:str
        One of 'urban', 'clean' or 'oilsand'
    psf:float
        The FWHM point spread function used to calculate radiances. One of '0.2, 0.4, 0.6, 0.8 or 1.0
    albedo:float
        The albedo used to calculate radiances. One of 0.0, 0.3 or 1.0
    year_month:str
        The day of the year used to calculate radiances. One of 2018_12, 2019_03, 2019_06 or 2019_09
    hours:List[int]
        The hours in teh day at which radiances are calculated. Typically [0,2,4,6,8,10,12,14,16,18,20,22]
    snr_range:float
        The range of signal to noise values to consider, typically 500 to 2000 in user defined steps
    numsamples: int
        The number of samples used to
    :return:
    """

    outputname = make_stats_path( albedo=albedo,profile=profile, psf=psf, year_month=year_month)
    if not os.path.exists(outputname):


        hourly_answers = {}
        for hour in hours:
            print(f'Processing {profile} psf = {psf} albedo = {albedo} for {year_month}-21 {hour}:00:00')

            path           = make_path(albedo=albedo,profile=profile, psf=psf, hour=hour, year_month=year_month)
            measurement_l1 = Radiance_AimNorthStudyMar2019(path)

            # create SCD object
            instrument     = Gaussian(fwhm=float(psf))
            cross_sections = CrossSectionOMNO2(instrument)
            solar_spectrum = SolarSpectrumOMNO2(instrument)
            ring_spectrum  = RingSpectrumOMNO2(solar_spectrum)
            scd_object     = SlantColumnOMNO2(cross_sections, ring_spectrum)

            # create AMF object
            amf_atmosphere = Atmosphere_AimNorthStudyMar2019()
            amf_lookup     = AMFLookup_AimNorthStudyMar2019(grid=5e2, cache_file='aimnorth_study_mar2019_no2_amf_cache.pkl')
            amf_object     = AMF_AimNorthStudyMar2019(amf_lookup, amf_atmosphere)

            retrieval      = RetrievalOMNO2(scd_object, amf_object)

            radiance = measurement_l1.radiance()
            nlos     = radiance.shape[1]

            snr_answers = {}
            for snr in snr_range:
                stats               = {}
                sd_error            = radiance/snr
                measurement_l1._err = sd_error

                stats['vcd']        = np.zeros([nlos, numsamples])
                stats['vcd_error']  = np.zeros([nlos, numsamples])
                stats['scd']        = np.zeros([nlos, numsamples])
                stats['scd_error']  = np.zeros([nlos, numsamples])
                stats['amf']        = np.zeros([nlos, numsamples])
                stats['amf_error']  = np.zeros([nlos, numsamples])

                for i in range(numsamples):
                    measurement_l1._rad = np.random.normal( radiance, sd_error)
                    results = retrieval.retrieve(measurement_l1)
                    stats['vcd'][:,i]       = results['vcd']
                    stats['vcd_error'][:,i] = results['vcd_error']
                    stats['scd'][:,i]       = results['scd']
                    stats['scd_error'][:,i] = results['scd_error']
                    stats['amf'][:,i]       = results['amf']
                    stats['amf_error'][:,i] = results['amf_error']

                # get true VCDs and calculate difference
                shells                   = amf_lookup.altitude_edges()
                true_profile             = amf_atmosphere.profile_property('no2', measurement_l1, shells)[0, :]
                stats['true_shells']     = shells
                stats['true_no2profile'] = true_profile
                stats['true_vcd']        = 1e2 * scipy.integrate.trapz( true_profile, shells)

                snr_answers[snr] = stats
            hourly_answers[hour] = snr_answers

        numhours = len(hours)
        numsnr   = len(snr_range)
        sorted_answers = {}
        for snr in snr_range:
            daily                    = {}
            daily['snr']             = snr
            daily['hours']           = hours
            daily['day']             = year_month
            daily['psf']             = psf
            daily['albedo']          = albedo
            daily['profile_type']    = profile
            daily['vcd']             = np.zeros( [nlos, numsamples, numhours])
            daily['vcd_error']       = np.zeros( [nlos, numsamples, numhours])
            daily['scd']             = np.zeros( [nlos, numsamples, numhours])
            daily['scd_error']       = np.zeros( [nlos, numsamples, numhours])
            daily['amf']             = np.zeros( [nlos, numsamples, numhours])
            daily['amf_error']        = np.zeros( [nlos, numsamples, numhours])
            for ih in range(numhours):
                snr_answers                = hourly_answers[hours[ih]]
                stats                      = snr_answers[snr]
                daily['vcd']      [:,:,ih] = stats['vcd']
                daily['vcd_error'][:,:,ih] = stats['vcd_error']
                daily['scd']      [:,:,ih] = stats['scd']
                daily['scd_error'][:,:,ih] = stats['scd_error']
                daily['amf']      [:,:,ih] = stats['amf']
                daily['amf_error'][:,:,ih] = stats['amf_error']

            daily['true_shells']     = stats['true_shells']
            daily['true_no2profile'] = stats['true_no2profile']
            daily['true_vcd']        = stats['true_vcd']

            sorted_answers[snr] = daily
        print("Generated the sorted answers")
        outputdir,name = os.path.split(outputname)
        try:
            os.makedirs(outputdir, exist_ok=True)
        except:
            print("Unable to create directory {}:".format(outputdir), sys.exc_info()[0])
        try:
            with open(outputname, 'wb') as f:  # and save the results to a pickle file
                pickle.dump(sorted_answers, f, pickle.HIGHEST_PROTOCOL)
        except:
            print("Unexpected file I/O error:", sys.exc_info()[0])

    #pdiff    = 1e2 * result['vcd'] / true_vcd - 1e2
        #
        # print(f'retrieved VCD: {np.average(result["vcd"]):.4e}')
        # print(f'true VCD:      {true_vcd:.4e}')
        # print(f'% difference:  {(np.average(pdiff)):.4f}')


if __name__ == '__main__':
    # define full range of scenarios
    albedos     = [ 0.3, 1.0, 0.0]
    year_months = ['2018-12', '2019-03', '2019-06', '2019-09']
    profiles    = ['clean', 'urban', 'oilsand']
    psfs        = [ 0.4, 0.8, 0.6, 1.0, 0.2]
    hours       = [ 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22]
    snr_range   = range(500,2101,200)
    numsamples  = 100


    #hours = [12,14]


    for profile in profiles:
        for psf in psfs:
            for albedo in albedos:
                for year_month in year_months:
                    do_retrievals( profile, psf, albedo, year_month, hours, snr_range, numsamples)
    print('Done')